import os
import logging
from collections import defaultdict

from fs.errors import ResourceNotFound
from lml.plugin import PluginManager

from moban import constants, exceptions
from moban.core import utils
from moban.externals import reporter, file_system
from moban.deprecated import deprecated_moban_path_notation
from moban.core.context import Context
from moban.core.strategy import Strategy
from moban.core.hashstore import HASH_STORE
from moban.externals.buffered_writer import BufferedWriter

LOG = logging.getLogger(__name__)


class MobanFactory(PluginManager):
    def __init__(self):
        super(MobanFactory, self).__init__(constants.TEMPLATE_ENGINE_EXTENSION)
        self.extensions = defaultdict(set)
        self.options_registry = {}

    def register_extensions(self, extensions):
        for user_template_type in extensions.keys():
            template_type = self.get_primary_key(user_template_type)

            LOG.debug(
                "Registering extensions: {0}={1}".format(
                    user_template_type, extensions[user_template_type]
                )
            )
            if template_type in self.extensions:
                self.extensions[template_type] = self.extensions[
                    user_template_type
                ].union(extensions[user_template_type])
            else:
                self.extensions[template_type] = extensions[user_template_type]

    def register_options(self, template_types):
        # need the value of 'template_types'
        # see test_get_user_defined_engine for help
        self.options_registry.update(template_types)

    def get_engine(self, template_type, template_dirs, context_dirs):
        template_dirs = list(expand_template_directories(template_dirs))
        template_dirs = utils.verify_the_existence_of_directories(
            template_dirs
        )
        if template_type in self.options_registry:

            custom_engine_spec = self.options_registry[template_type]
            engine_cls = self.load_me_now(
                custom_engine_spec[constants.TEMPLATE_TYPES_BASE_TYPE]
            )
            options = custom_engine_spec[constants.TEMPLATE_TYPES_OPTIONS]
        else:
            engine_cls = self.load_me_now(template_type)
            engine_extensions = self.extensions.get(template_type)
            if engine_extensions:
                options = dict(extensions=list(engine_extensions))
            else:
                options = dict()
        template_fs = file_system.get_multi_fs(template_dirs)
        engine = engine_cls(template_fs, options)
        return MobanEngine(template_fs, context_dirs, engine)

    def get_primary_key(self, template_type):
        for key, item in self.options_registry.items():
            if template_type in item[constants.TEMPLATE_TYPES_FILE_EXTENSIONS]:
                return key

        return super(MobanFactory, self).get_primary_key(template_type)

    def all_types(self):
        return list(self.registry.keys()) + list(self.options_registry.keys())

    def raise_exception(self, key):
        raise exceptions.NoThirdPartyEngine(key)


class MobanEngine(object):
    def __init__(self, template_fs, context_dirs, engine):
        context_dirs = expand_template_directory(context_dirs)
        self.context = Context(context_dirs)
        self.template_fs = template_fs
        self.engine = engine
        self.templated_count = 0
        self.file_count = 0
        self.buffered_writer = BufferedWriter()
        self.engine_action = getattr(
            engine,
            "ACTION_IN_PRESENT_CONTINUOUS_TENSE",
            constants.LABEL_MOBAN_ACTION_IN_PRESENT_CONTINUOUS_TENSE,
        )
        self.engine_actioned = getattr(
            engine,
            "ACTION_IN_PAST_TENSE",
            constants.LABEL_MOBAN_ACTION_IN_PAST_TENSE,
        )

    def report(self):
        if self.templated_count == 0:
            reporter.report_no_action()
        elif self.templated_count == self.file_count:
            reporter.report_full_run(self.engine_actioned, self.file_count)
        else:
            reporter.report_partial_run(
                self.engine_actioned, self.templated_count, self.file_count
            )

    def number_of_templated_files(self):
        return self.templated_count

    def render_to_file(self, template_file, data_file, output_file):
        data = self.context.get_data(data_file)
        template = self.engine.get_template(template_file)
        try:
            template_abs_path = self.template_fs.geturl(
                template_file, purpose="fs"
            )
        except ResourceNotFound:
            template_abs_path = template_file

        flag = self.apply_template(
            template_abs_path, template, data, output_file
        )
        if flag:
            reporter.report_templating(
                self.engine_action, template_file, output_file
            )
            self.templated_count += 1
        self.file_count += 1

        self.buffered_writer.close()

    def render_string_to_file(
        self, template_in_string, data_file, output_file
    ):
        template = self.engine.get_template_from_string(template_in_string)
        template_abs_path = f"{template_in_string[:10]}..."
        data = self.context.get_data(data_file)
        flag = self.apply_template(
            template_abs_path, template, data, output_file
        )
        if flag:
            reporter.report_templating(
                self.engine_action, template_abs_path, output_file
            )
            self.templated_count += 1
        self.file_count += 1
        self.buffered_writer.close()

    def apply_template(self, template_abs_path, template, data, output_file):
        rendered_content = self.engine.apply_template(
            template, data, output_file
        )
        if not isinstance(rendered_content, bytes):
            rendered_content = rendered_content.encode("utf-8")

        try:
            flag = HASH_STORE.is_file_changed(
                output_file, rendered_content, template_abs_path
            )
            if flag:
                self.buffered_writer.write_file_out(
                    output_file, rendered_content
                )
                if not file_system.is_zip_alike_url(output_file):
                    try:
                        file_system.file_permissions_copy(
                            template_abs_path, output_file
                        )
                    except exceptions.NoPermissionsNeeded:
                        # HttpFs does not have getsyspath
                        # zip, tar have no permission
                        # win32 does not work
                        pass
            return flag
        except exceptions.FileNotFound:
            # the template is a string from command line
            LOG.info(f"{template_abs_path} is not a file")
            self.buffered_writer.write_file_out(output_file, rendered_content)
            return True

    def render_to_files(self, array_of_template_targets):
        sta = Strategy(array_of_template_targets)

        sta.process()
        choice = sta.what_to_do()
        if choice == Strategy.DATA_FIRST:
            self._render_with_finding_data_first(sta.data_file_index)
        else:
            self._render_with_finding_template_first(sta.template_file_index)
        self.buffered_writer.close()

    def _render_with_finding_template_first(self, template_file_index):
        for (template_file, data_output_pairs) in template_file_index.items():
            template = self.engine.get_template(template_file)
            template_abs_path = self.template_fs.geturl(
                template_file, purpose="fs"
            )
            for (data_file, output) in data_output_pairs:
                data = self.context.get_data(data_file)
                flag = self.apply_template(
                    template_abs_path, template, data, output
                )
                if flag:
                    reporter.report_templating(
                        self.engine_action, template_file, output
                    )
                    self.templated_count += 1
                self.file_count += 1

    def _render_with_finding_data_first(self, data_file_index):
        for (data_file, template_output_pairs) in data_file_index.items():
            data = self.context.get_data(data_file)
            for (template_file, output) in template_output_pairs:
                template = self.engine.get_template(template_file)
                if isinstance(template, bool):
                    if template:
                        reporter.report_templating(
                            self.engine_action, template_file, None
                        )
                        self.templated_count += 1
                else:
                    template_abs_path = self.template_fs.geturl(
                        template_file, purpose="fs"
                    )
                    flag = self.apply_template(
                        template_abs_path, template, data, output
                    )
                    if flag:
                        reporter.report_templating(
                            self.engine_action, template_file, output
                        )
                        self.templated_count += 1
                self.file_count += 1


def expand_template_directories(dirs):
    LOG.debug(f"Expanding {dirs}...")
    if not isinstance(dirs, list):
        dirs = [dirs]

    for directory in dirs:
        yield expand_template_directory(directory)


def expand_template_directory(directory):
    LOG.debug(f"Expanding {directory}...")
    translated_directory = None
    if ":" in directory and directory[1] != ":" and "://" not in directory:
        translated_directory = deprecated_moban_path_notation(directory)
    elif "://" in directory:
        translated_directory = directory
    else:
        # local template path
        translated_directory = os.path.normcase(os.path.abspath(directory))
        translated_directory = file_system.fs_url(translated_directory)
    return translated_directory



import uuid

import fs.path
from nose.tools import eq_, raises

from moban.exceptions import GroupTargetNotFound
from moban.core.mobanfile import targets
from moban.core.definitions import TemplateTarget

TEMPLATE = "a.jj2"
OUTPUT = "a.output"
CONFIGURATION = "data.config"
TEMPLATE_DIRS = [fs.path.join("tests", "fixtures")]
DEFAULT_TEMPLATE_TYPE = "default-template-type"


def test_handling_group_target():
    group_template_type = "group-template-type"
    options = dict(
        configuration=CONFIGURATION,
        template_type=DEFAULT_TEMPLATE_TYPE,
        template_dir=TEMPLATE_DIRS,
    )
    short_hand_targets = [{OUTPUT: TEMPLATE}]

    actual = list(
        targets._handle_group_target(
            options, short_hand_targets, group_template_type
        )
    )
    expected = [
        TemplateTarget(TEMPLATE, CONFIGURATION, OUTPUT, group_template_type)
    ]
    eq_(expected, actual)


def test_extract_group_targets():
    test_targets = [
        {"output": "a.output", "template": "a.template.jj2"},
        {"copy": [{"output": "source"}], "copy1": [{"output1": "source1"}]},
    ]
    actual = targets.extract_group_targets("copy1", test_targets)
    expected = [{"copy1": [{"output1": "source1"}]}]
    eq_(expected, actual)


@raises(GroupTargetNotFound)
def test_extract_group_targets_not_found():
    test_targets = [
        {"copy": [{"output": "source"}], "copy1": [{"output1": "source1"}]}
    ]
    actual = targets.extract_group_targets("copy2", test_targets)
    expected = []
    eq_(expected, actual)


class TestImplicitTarget:
    def test_derive_template_type_from_target_template_file(self):

        options = dict(
            configuration=CONFIGURATION,
            template_type=DEFAULT_TEMPLATE_TYPE,
            template_dir=TEMPLATE_DIRS,
        )

        actual = list(
            targets._handle_implicit_target(options, TEMPLATE, OUTPUT)
        )
        expected = [TemplateTarget(TEMPLATE, CONFIGURATION, OUTPUT, "jj2")]
        eq_(expected, actual)

    def test_use_moban_default_template_from_options(self):
        template_without_suffix = "template"

        options = dict(
            configuration=CONFIGURATION,
            template_type=DEFAULT_TEMPLATE_TYPE,
            template_dir=TEMPLATE_DIRS,
        )

        actual = list(
            targets._handle_implicit_target(
                options, template_without_suffix, OUTPUT
            )
        )
        expected = [
            TemplateTarget(
                template_without_suffix,
                CONFIGURATION,
                OUTPUT,
                DEFAULT_TEMPLATE_TYPE,
            )
        ]
        eq_(expected, actual)


class TestExplicitTarget:
    def test_use_target_template_type(self):

        target = dict(template_type="use-me", template=TEMPLATE, output=OUTPUT)
        options = dict(
            configuration=CONFIGURATION,
            template_type=DEFAULT_TEMPLATE_TYPE,
            template_dir=TEMPLATE_DIRS,
        )

        actual = list(targets._handle_explicit_target(options, target))
        expected = [TemplateTarget(TEMPLATE, CONFIGURATION, OUTPUT, "use-me")]
        eq_(expected, actual)

    def test_derive_template_type_from_target_template_file(self):

        target = dict(template=TEMPLATE, output=OUTPUT)
        options = dict(
            configuration=CONFIGURATION,
            template_type=DEFAULT_TEMPLATE_TYPE,
            template_dir=TEMPLATE_DIRS,
        )

        actual = list(targets._handle_explicit_target(options, target))
        expected = [TemplateTarget(TEMPLATE, CONFIGURATION, OUTPUT, "jj2")]
        eq_(expected, actual)

    def test_use_moban_default_template_from_options(self):
        template_without_suffix = "template"
        target = dict(template=template_without_suffix, output=OUTPUT)
        options = dict(
            configuration=CONFIGURATION,
            template_type=DEFAULT_TEMPLATE_TYPE,
            template_dir=TEMPLATE_DIRS,
        )

        actual = list(targets._handle_explicit_target(options, target))
        expected = [
            TemplateTarget(
                template_without_suffix,
                CONFIGURATION,
                OUTPUT,
                DEFAULT_TEMPLATE_TYPE,
            )
        ]
        eq_(expected, actual)

    def test_ad_hoc_type(self):
        target = dict(template=TEMPLATE, output=OUTPUT)
        template_type = [
            {"base_type": "jinja2"},
            {
                "options": [
                    {"block_end_string": "*))"},
                    {"block_start_string": "((*"},
                ]
            },
        ]
        options = dict(
            configuration=CONFIGURATION,
            template_type=template_type,
            template_dir=TEMPLATE_DIRS,
        )

        actual = list(targets._handle_explicit_target(options, target))
        file_extension = uuid.uuid4().hex
        expected = [
            TemplateTarget(TEMPLATE, CONFIGURATION, OUTPUT, file_extension)
        ]
        eq_(actual, expected)



from mock import patch
from nose.tools import eq_

from moban.deprecated import GitRequire


@patch("moban.deprecated.pip_install")
def test_handle_requires_pypkg(fake_pip_install):
    modules = ["package1", "package2"]
    from moban.deprecated import handle_requires

    handle_requires(modules)
    fake_pip_install.assert_called_with(modules)


@patch("moban.deprecated.pip_install")
def test_handle_requires_pypkg_with_alternative_syntax(fake_pip_install):
    modules = [{"type": "pypi", "name": "pypi-mobans"}]
    from moban.mobanfile import handle_requires

    handle_requires(modules)
    fake_pip_install.assert_called_with(["pypi-mobans"])


@patch("moban.deprecated.git_clone")
def test_handle_requires_repos(fake_git_clone):
    repos = ["https://github.com/my/repo", "https://gitlab.com/my/repo"]
    from moban.mobanfile import handle_requires

    expected = []
    for repo in repos:
        expected.append(GitRequire(git_url=repo, submodule=False))

    handle_requires(repos)
    fake_git_clone.assert_called_with(expected)


@patch("moban.deprecated.git_clone")
def test_handle_requires_repos_with_alternative_syntax(fake_git_clone):
    repos = [{"type": "git", "url": "https://github.com/my/repo"}]
    from moban.mobanfile import handle_requires

    handle_requires(repos)
    fake_git_clone.assert_called_with(
        [GitRequire(git_url="https://github.com/my/repo")]
    )


@patch("moban.deprecated.pip_install")
@patch("moban.deprecated.git_clone")
def test_handle_requires_repos_with_submodule(
    fake_git_clone, fake_pip_install
):
    repos = [
        {"type": "git", "url": "https://github.com/my/repo", "submodule": True}
    ]
    from moban.mobanfile import handle_requires

    handle_requires(repos)
    fake_git_clone.assert_called_with(
        [GitRequire(git_url="https://github.com/my/repo", submodule=True)]
    )
    eq_(fake_pip_install.called, False)


def test_is_repo():
    repos = [
        "https://github.com/my/repo",
        "https://gitlab.com/my/repo",
        "https://bitbucket.com/my/repo",
        "https://unsupported.com/my/repo",
        "invalid/repo/url",
    ]
    from moban.deprecated import is_repo

    actual = [is_repo(repo) for repo in repos]
    expected = [True, True, True, False, False]
    eq_(expected, actual)



import os
import sys

import fs.path
from mock import patch
from lml.plugin import PluginInfo
from nose.tools import eq_, raises

import moban.exceptions as exceptions
from moban.core import ENGINES
from moban.core.context import Context
from moban.jinja2.engine import (
    Engine,
    is_extension_list_valid,
    import_module_of_extension,
)
from moban.core.moban_factory import MobanEngine, expand_template_directories

USER_HOME = fs.path.join("user", "home", ".moban", "repos")


@PluginInfo("library", tags=["testmobans"])
class TestPypkg:
    def __init__(self):
        __package_path__ = os.path.normcase(os.path.dirname(__file__))
        self.resources_path = os.path.join(__package_path__, "fixtures")


def test_expand_pypi_dir():
    dirs = list(expand_template_directories("testmobans:template-tests"))
    for directory in dirs:
        assert os.path.exists(directory)


@patch("moban.deprecated.repo.get_moban_home", return_value=USER_HOME)
@patch("moban.file_system.exists", return_value=True)
def test_expand_repo_dir(_, __):
    dirs = list(expand_template_directories("git_repo:template"))

    expected = [fs.path.join(USER_HOME, "git_repo", "template")]
    eq_(expected, dirs)


def test_default_template_type():
    engine = ENGINES.get_engine("jj2", [], "")
    assert engine.engine.__class__ == Engine


class FakeEngine:
    def __init__(self, template_dirs, extensions=None):
        pass


@patch("lml.plugin.PluginManager.load_me_now", return_value=FakeEngine)
def test_default_mako_type(_):  # fake mako
    engine = ENGINES.get_engine("fake", [], "")
    assert engine.engine.__class__ == FakeEngine


@raises(exceptions.NoThirdPartyEngine)
def test_unknown_template_type():
    ENGINES.get_engine("unknown_template_type", [], "")


@raises(exceptions.DirectoryNotFound)
def test_non_existent_tmpl_directries():
    ENGINES.get_engine("jj2", "idontexist", "")


@raises(exceptions.DirectoryNotFound)
def test_non_existent_config_directries():
    MobanEngine("tests", "abc", Engine)


@raises(exceptions.DirectoryNotFound)
def test_non_existent_ctx_directries():
    Context(["abc"])


def test_file_tests():
    output = "test.txt"
    path = fs.path.join("tests", "fixtures", "jinja_tests")
    engine = ENGINES.get_engine("jinja2", [path], path)
    engine.render_to_file("file_tests.template", "file_tests.yml", output)
    with open(output, "r") as output_file:
        content = output_file.read()
        eq_(content, "yes\nhere")
    eq_(engine.file_count, 1)
    eq_(engine.templated_count, 1)
    os.unlink(output)


def test_render_string_to_file():
    output = "test.txt"
    path = fs.path.join("tests", "fixtures", "jinja_tests")
    engine = ENGINES.get_engine("jinja2", [path], path)
    engine.render_string_to_file("{{test}}", "file_tests.yml", output)
    with open(output, "r") as output_file:
        content = output_file.read()
        eq_(content, "here")
    eq_(engine.file_count, 1)
    eq_(engine.templated_count, 1)
    os.unlink(output)


def test_global_template_variables():
    output = "test.txt"
    path = fs.path.join("tests", "fixtures", "globals")
    engine = ENGINES.get_engine("jinja2", [path], path)
    engine.render_to_file("variables.template", "variables.yml", output)
    with open(output, "r") as output_file:
        content = output_file.read()
        eq_(content, "template: variables.template\ntarget: test.txt\nhere")
    os.unlink(output)


def test_nested_global_template_variables():
    output = "test.txt"
    path = fs.path.join("tests", "fixtures", "globals")
    engine = ENGINES.get_engine("jinja2", [path], path)
    engine.render_to_file("nested.template", "variables.yml", output)
    with open(output, "r") as output_file:
        content = output_file.read()
        eq_(content, "template: nested.template\ntarget: test.txt\nhere")
    os.unlink(output)


def test_environ_variables_as_data():
    test_var = "TEST_ENVIRONMENT_VARIABLE"
    test_value = "foo"
    os.environ[test_var] = test_value
    output = "test.txt"
    path = fs.path.join("tests", "fixtures", "environ_vars_as_data")
    engine = ENGINES.get_engine("jinja2", [path], path)
    engine.render_to_file("test.template", "this_does_not_exist.yml", output)
    with open(output, "r") as output_file:
        content = output_file.read()
        eq_(content, "foo")
    os.unlink(output)


def test_string_template():
    output = "test.txt"
    path = fs.path.join("tests", "fixtures")
    engine = ENGINES.get_engine("jinja2", [path], path)
    engine.render_string_to_file("{{simple}}", "simple.yaml", output)
    with open(output, "r") as output_file:
        content = output_file.read()
        eq_(content, "yaml")
    os.unlink(output)


def test_extensions_validator():
    test_fixtures = [None, ["module1", "module2"], []]
    expected = [False, True, False]
    actual = []
    for fixture in test_fixtures:
        actual.append(is_extension_list_valid(fixture))

    eq_(expected, actual)


def test_import():
    extensions = [
        "jinja2.ext.do",
        "jinja2_time.TimeExtension",
        "jinja2.ext.loopcontrols",
    ]
    import_module_of_extension(extensions)
    modules = ["jinja2", "jinja2_time"]
    for module in modules:
        assert module in sys.modules



import os

import fs.path
from mock import patch
from nose.tools import eq_

from moban.core import ENGINES
from moban.definitions import TemplateTarget
from moban.jinja2.engine import Engine
from moban.data_loaders.yaml import open_yaml

MODULE = "moban.core.moban_factory"


@patch(MODULE + ".MobanEngine._render_with_finding_data_first")
def test_do_templates_1(_do_templates_with_more_shared_data):
    jobs = [
        TemplateTarget("1.template", "data.yml", "1.output"),
        TemplateTarget("2.template", "data.yml", "2.output"),
        TemplateTarget("3.template", "data.yml", "3.output"),
        TemplateTarget("4.template", "data.yml", "4.output"),
        TemplateTarget("5.template", "data.yml", "6.output"),
    ]
    expected = {
        "data.yml": [
            ("1.template", "1.output"),
            ("2.template", "2.output"),
            ("3.template", "3.output"),
            ("4.template", "4.output"),
            ("5.template", "6.output"),
        ]
    }
    engine = ENGINES.get_engine("jinja2", ".", ".")
    engine.render_to_files(jobs)
    _do_templates_with_more_shared_data.assert_called_with(expected)


@patch(MODULE + ".MobanEngine._render_with_finding_template_first")
def test_do_templates_2(_do_templates_with_more_shared_templates):
    jobs = [
        TemplateTarget("1.template", "data1.yml", "1.output"),
        TemplateTarget("1.template", "data2.yml", "2.output"),
        TemplateTarget("1.template", "data3.yml", "3.output"),
        TemplateTarget("1.template", "data4.yml", "4.output"),
        TemplateTarget("1.template", "data5.yml", "6.output"),
    ]
    expected = {
        "1.template": [
            ("data1.yml", "1.output"),
            ("data2.yml", "2.output"),
            ("data3.yml", "3.output"),
            ("data4.yml", "4.output"),
            ("data5.yml", "6.output"),
        ]
    }
    engine = ENGINES.get_engine("jinja2", ".", ".")
    engine.render_to_files(jobs)
    _do_templates_with_more_shared_templates.assert_called_with(expected)


def test_do_templates_with_more_shared_templates():
    base_dir = fs.path.join("tests", "fixtures")
    engine = ENGINES.get_engine(
        "jinja2", base_dir, fs.path.join(base_dir, "config")
    )
    engine._render_with_finding_template_first(
        {"a.jj2": [(fs.path.join(base_dir, "child.yaml"), "test")]}
    )
    with open("test", "r") as f:
        content = f.read()
        assert content == "hello world ox"
    os.unlink("test")


def test_do_templates_with_more_shared_data():
    base_dir = fs.path.join("tests", "fixtures")
    engine = ENGINES.get_engine(
        "jinja2", base_dir, fs.path.join(base_dir, "config")
    )
    engine._render_with_finding_data_first(
        {fs.path.join(base_dir, "child.yaml"): [("a.jj2", "test")]}
    )
    with open("test", "r") as f:
        content = f.read()
        assert content == "hello world ox"
    os.unlink("test")


def test_get_user_defined_engine():
    test_fixture = fs.path.join(
        "tests", "fixtures", "mobanengine", "sample_template_type.yml"
    )
    template_types = open_yaml(test_fixture)
    ENGINES.register_options(template_types["template_types"])
    engine = ENGINES.get_engine("custom_jinja", ".", ".")
    eq_(engine.engine.__class__, Engine)


def test_custom_file_extension_is_assocated_with_user_defined_engine():
    test_fixture = fs.path.join(
        "tests", "fixtures", "mobanengine", "sample_template_type.yml"
    )
    template_types = open_yaml(test_fixture)
    ENGINES.register_options(template_types["template_types"])
    template_type = ENGINES.get_primary_key("demo_file_suffix")
    eq_("custom_jinja", template_type)


def test_built_in_jinja2_file_extension_still_works():
    test_fixture = fs.path.join(
        "tests", "fixtures", "mobanengine", "sample_template_type.yml"
    )
    template_types = open_yaml(test_fixture)
    ENGINES.register_options(template_types["template_types"])
    template_type = ENGINES.get_primary_key("jj2")
    eq_("jinja2", template_type)



import logging

from moban import constants
from moban.externals import file_system

LOG = logging.getLogger(__name__)


def verify_the_existence_of_directories(dirs):
    if not isinstance(dirs, list):
        dirs = [dirs]

    dirs = [
        directory
        for directory in dirs
        if not (
            constants.DEFAULT_CONFIGURATION_DIRNAME in directory
            or constants.DEFAULT_TEMPLATE_DIRNAME in directory
        )
    ]
    if file_system.exists(constants.DEFAULT_CONFIGURATION_DIRNAME):
        dirs.append(constants.DEFAULT_CONFIGURATION_DIRNAME)

    if file_system.exists(constants.DEFAULT_TEMPLATE_DIRNAME):
        dirs.append(constants.DEFAULT_TEMPLATE_DIRNAME)
    return dirs



import uuid
import logging

from moban import core, constants, exceptions
from moban.externals import reporter
from moban.core.definitions import TemplateTarget
from .store import STORE
from .templates import handle_template

LOG = logging.getLogger(__name__)


def extract_target(options):
    template, config, output = (
        options.get(constants.LABEL_TEMPLATE),
        options.get(constants.LABEL_CONFIG),
        options.get(constants.LABEL_OUTPUT),
    )
    result = []
    if template:
        if output is None:
            raise Exception(
                f"Please specify a output file name for {template}"
            )
        if config:
            result = {
                constants.LABEL_TEMPLATE: template,
                constants.LABEL_CONFIG: config,
                constants.LABEL_OUTPUT: output,
            }

        else:
            result = {output: template}
    return result


def extract_group_targets(group, targets):
    for target in targets:
        if constants.LABEL_OUTPUT in target:
            continue

        for group_name, group_targets in target.items():
            if isinstance(group_targets, str) is False and group_name == group:
                # grouping by template type feature
                return [{group_name: group_targets}]
    raise exceptions.GroupTargetNotFound(f"{group} is not found")


def parse_targets(options, targets):
    LOG.info("paring targets..")
    STORE.init()
    for target in targets:
        if constants.LABEL_OUTPUT in target:
            for template_target in _handle_explicit_target(options, target):
                STORE.add(template_target)
        else:
            for output, template_file in target.items():
                if isinstance(template_file, str):
                    for template_target in _handle_implicit_target(
                        options, template_file, output
                    ):
                        STORE.add(template_target)
                else:
                    # grouping by template type feature
                    group_template_type = output
                    a_list_short_hand_targets = template_file
                    for template_target in _handle_group_target(
                        options, a_list_short_hand_targets, group_template_type
                    ):
                        STORE.add(template_target)


def _handle_explicit_target(options, target):
    common_data_file = options[constants.LABEL_CONFIG]
    default_template_type = options[constants.LABEL_TEMPLATE_TYPE]
    template_file = target.get(
        constants.LABEL_TEMPLATE, options.get(constants.LABEL_TEMPLATE, None)
    )
    data_file = target.get(constants.LABEL_CONFIG, common_data_file)
    output = target[constants.LABEL_OUTPUT]
    if output:
        template_type = target.get(constants.LABEL_TEMPLATE_TYPE)
    else:
        template_type = constants.TEMPLATE_DELETE
    if template_type and len(template_type) > 0:
        if constants.TEMPLATE_TYPES_FILE_EXTENSIONS in template_type:
            reporter.report_file_extension_not_needed()
        if constants.TEMPLATE_TYPES_BASE_TYPE in template_type:
            adhoc_type = uuid.uuid4().hex
            file_extension = uuid.uuid4().hex
            base_type = template_type[constants.TEMPLATE_TYPES_BASE_TYPE]
            template_types_options = template_type[
                constants.TEMPLATE_TYPES_OPTIONS
            ]
            the_adhoc_type = {
                adhoc_type: {
                    constants.TEMPLATE_TYPES_FILE_EXTENSIONS: [file_extension],
                    constants.TEMPLATE_TYPES_BASE_TYPE: base_type,
                    constants.TEMPLATE_TYPES_OPTIONS: template_types_options,
                }
            }
            core.ENGINES.register_options(the_adhoc_type)
            template_type = file_extension
    for src, dest, t_type in handle_template(
        template_file, output, options[constants.LABEL_TMPL_DIRS]
    ):
        if template_type:
            yield TemplateTarget(src, data_file, dest, template_type)
        else:
            if t_type:
                yield TemplateTarget(src, data_file, dest, t_type)
            else:
                yield TemplateTarget(
                    src, data_file, dest, default_template_type
                )


def _handle_group_target(
    options, a_list_short_hand_targets, group_template_type
):
    # grouping by template type feature
    common_data_file = options[constants.LABEL_CONFIG]
    for _output, _template_file in _iterate_list_of_dicts(
        a_list_short_hand_targets
    ):
        for src, dest, t_type in handle_template(
            _template_file, _output, options[constants.LABEL_TMPL_DIRS]
        ):
            yield TemplateTarget(
                src, common_data_file, dest, group_template_type
            )


def _handle_implicit_target(options, template_file, output):
    common_data_file = options[constants.LABEL_CONFIG]
    default_template_type = options[constants.LABEL_TEMPLATE_TYPE]
    for src, dest, t_type in handle_template(
        template_file, output, options[constants.LABEL_TMPL_DIRS]
    ):
        if t_type:
            yield TemplateTarget(src, common_data_file, dest, t_type)
        else:
            yield TemplateTarget(
                src, common_data_file, dest, default_template_type
            )


def _iterate_list_of_dicts(list_of_dict):
    for adict in list_of_dict:
        for key, value in adict.items():
            yield (key, value)



import os
import tempfile

from nose.tools import eq_

from moban.externals import file_system
from moban.externals.buffered_writer import BufferedWriter, write_file_out

CONTENT = b"""
    helloworld




    """
EXPECTED = "\n    helloworld\n\n\n\n\n    "


class TestBufferedWriter:
    def setUp(self):
        self.writer = BufferedWriter()

    def test_write_text(self):
        test_file = "testout"
        self.writer.write_file_out(test_file, CONTENT)
        self.writer.close()
        content = file_system.read_text(test_file)
        eq_(content, EXPECTED)
        os.unlink(test_file)

    def test_write_a_zip(self):
        tmp_dir = os.path.normcase(tempfile.gettempdir())
        test_file = "zip://" + tmp_dir + "/testout.zip!/testout"
        self.writer.write_file_out(test_file, CONTENT)
        self.writer.close()
        content = file_system.read_text(test_file)
        eq_(content, EXPECTED)
        os.unlink(os.path.join(tmp_dir, "testout.zip"))


def test_write_file_out():
    test_file = "testout"
    write_file_out(test_file, CONTENT)
    with open(test_file, "r") as f:
        content = f.read()
        eq_(content, EXPECTED)



import sys

from mock import patch
from nose.tools import eq_

from moban.externals import reporter

PY2 = sys.version_info[0] == 2
if PY2:
    from StringIO import StringIO
else:
    from io import StringIO


def test_partial_run():
    patcher = patch("sys.stdout", new_callable=StringIO)
    fake_stdout = patcher.start()
    reporter.report_partial_run("Actioned", 1, 20)
    patcher.stop()
    eq_(fake_stdout.getvalue(), "Actioned 1 out of 20 files.\n")


def test_full_run():
    patcher = patch("sys.stdout", new_callable=StringIO)
    fake_stdout = patcher.start()
    reporter.report_full_run("Worked on", 20)
    patcher.stop()
    eq_(fake_stdout.getvalue(), "Worked on 20 files.\n")


def test_error_message():
    patcher = patch("sys.stdout", new_callable=StringIO)
    fake_stdout = patcher.start()
    reporter.report_error_message("something wrong")
    patcher.stop()
    eq_(fake_stdout.getvalue(), "Error: something wrong\n")


def test_info_message():
    patcher = patch("sys.stdout", new_callable=StringIO)
    fake_stdout = patcher.start()
    reporter.report_info_message("for your information")
    patcher.stop()
    eq_(fake_stdout.getvalue(), "Info: for your information\n")


def test_warning_message():
    patcher = patch("sys.stdout", new_callable=StringIO)
    fake_stdout = patcher.start()
    reporter.report_warning_message("Maybe you wanna know")
    patcher.stop()
    eq_(fake_stdout.getvalue(), "Warning: Maybe you wanna know\n")


def test_report_templating():
    patcher = patch("sys.stdout", new_callable=StringIO)
    fake_stdout = patcher.start()
    reporter.report_templating("Transforming", "a", "b")
    patcher.stop()
    eq_(fake_stdout.getvalue(), "Transforming a to b\n")


def test_no_action():
    patcher = patch("sys.stdout", new_callable=StringIO)
    fake_stdout = patcher.start()
    reporter.report_no_action()
    patcher.stop()
    eq_(fake_stdout.getvalue(), "No actions performed\n")


def test_format_single():
    message = "1 files"
    ret = reporter._format_single(message, 1)
    eq_(ret, "1 file")


def test_report_template_not_in_moban_file():
    patcher = patch("sys.stdout", new_callable=StringIO)
    fake_stdout = patcher.start()
    reporter.report_template_not_in_moban_file("test.jj2")
    patcher.stop()
    eq_(
        fake_stdout.getvalue(),
        "Warning: test.jj2 is not defined in your moban file!\n",
    )


def test_report_file_extension_not_needed():
    patcher = patch("sys.stdout", new_callable=StringIO)
    fake_stdout = patcher.start()
    reporter.report_file_extension_not_needed()
    patcher.stop()
    eq_(
        fake_stdout.getvalue(),
        "Info: File extension is not required for ad-hoc type\n",
    )



import os

from nose.tools import eq_

from moban import file_system
from moban.jinja2.engine import Engine
from moban.jinja2.extensions import jinja_global
from moban.core.moban_factory import MobanEngine


def test_globals():
    output = "globals.txt"
    test_dict = dict(hello="world")
    jinja_global("test", test_dict)
    path = os.path.join("tests", "fixtures", "globals")
    template_fs = file_system.get_multi_fs([path])
    engine = MobanEngine(template_fs, path, Engine(template_fs))
    engine.render_to_file("basic.template", "basic.yml", output)
    with open(output, "r") as output_file:
        content = output_file.read()
        eq_(content, "world\n\ntest")
    os.unlink(output)



import sys
from os.path import isabs, isdir, exists, isfile, islink, ismount, lexists

from moban.plugins.jinja2.extensions import jinja_tests

if sys.platform == "win32":
    from moban.jinja2.tests.win32 import samefile
else:
    from os.path import samefile

jinja_tests(
    is_dir=isdir,
    directory=isdir,
    is_file=isfile,
    file=isfile,
    is_link=islink,
    link=islink,
    exists=exists,
    link_exists=lexists,
    # path testing
    is_abs=isabs,
    abs=isabs,
    is_same_file=samefile,
    same_file=samefile,
    is_mount=ismount,
    mount=ismount,
)



import fs.path
from nose.tools import eq_

from moban.plugins.json_loader import open_json


def test_open_json():
    content = open_json(fs.path.join("tests", "fixtures", "child.json"))
    expected = {"key": "hello world", "pass": "ox"}
    eq_(expected, content)



import os

from nose.tools import eq_

from moban.main import load_engine_factory_and_engines
from moban.core.data_loader import load_data


def test_overrides_a_list_of_config_files():
    base_dir = os.path.join("tests", "fixtures", "issue_126")
    config_dir = os.path.join(base_dir, "config")
    actual = load_data(config_dir, os.path.join(base_dir, "the_config.yaml"))
    expected = [
        ("key", "value"),
        ("key_from_a", "apple"),
        ("key_from_b", "bee"),
    ]
    for item, expected_item in zip(actual.items(), expected):
        eq_(item, expected_item)

    eq_(len(actual), len(expected))


def test_overrides_a_list_of_config_files_but_cannot_find_them():
    base_dir = os.path.join("tests", "fixtures", "issue_126")
    actual = load_data(None, os.path.join(base_dir, "the_config.yaml"))

    expected = [("key", "value")]
    for item, expected_item in zip(actual.items(), expected):
        eq_(item, expected_item)

    eq_(len(actual), len(expected))


def test_overrides_ignores_override_sequence():
    base_dir = os.path.join("tests", "fixtures", "issue_126")
    config_dir = os.path.join(base_dir, "config")
    actual = load_data(config_dir, os.path.join(base_dir, "the_config.yaml"))
    expected = [
        ("key", "value"),
        ("key_from_a", "apple"),
        ("key_from_b", "bee"),
    ]
    for item, expected_item in zip(actual.items(), expected):
        eq_(item, expected_item)


def test_overrides_select_keys_from_parent_files():
    base_dir = os.path.join("tests", "fixtures", "issue_126")
    config_dir = os.path.join(base_dir, "config")
    actual = load_data(
        config_dir, os.path.join(base_dir, "multi-key-config.yaml")
    )
    expected = [
        ("cat", "from config"),
        ("alpha", "from a"),
        ("beta", "from b"),
    ]
    for item, expected_item in zip(actual.items(), expected):
        eq_(item, expected_item)


def test_overrides_select_keys():
    base_dir = os.path.join("tests", "fixtures", "issue_126")
    config_dir = os.path.join(base_dir, "config")
    actual = load_data(
        config_dir, os.path.join(base_dir, "multi-key-config-override.yaml")
    )
    expected = [
        ("alpha", "from config"),
        ("cat", "from config"),
        ("beta", "from b"),
    ]
    for item, expected_item in zip(actual.items(), expected):
        eq_(item, expected_item)


def test_overrides_nested_keys():
    base_dir = os.path.join("tests", "fixtures", "issue_126")
    config_dir = os.path.join(base_dir, "config")
    actual = load_data(config_dir, os.path.join(base_dir, "raspberry.yaml"))
    expected = {
        "raspberry": {
            "other": "OpenGL 3.0",
            "version": 4,
            "memory": "4GB",
            "core": "quad",
            "wifi": "2.5 & 5.0 GHz",
            "USB": 3.0,
            "Bluetooth": 5.0,
        },
        "tessel": {"version": 2, "USB": "micro", "wifi": "802.11gn"},
    }

    eq_(dict(actual), expected)


def test_overrides_fs_url():
    load_engine_factory_and_engines()
    base_dir = os.path.join("tests", "fixtures")
    actual = load_data(None, os.path.join(base_dir, "override_fs_url.yaml"))
    assert "requires" in actual



from nose.tools import eq_

from moban.jinja2.filters.text import split_length


def test_split_length():
    inputs = [
        ["some good issues are helping the developer for the", 12],
        ["http://github.com/chfw/abc is cool", 12],
        ["http://github.com/chfw/abc is cool", 100],
        ["some  extra     space will be     removed", 10],
    ]
    expectations = [
        ["some good", "issues are", "helping the", "developer", "for the"],
        ["http://github.com/chfw/abc", "is cool"],
        ["http://github.com/chfw/abc is cool"],
        ["some extra", "space will", "be removed"],
    ]
    for test, expect in zip(inputs, expectations):
        actual = split_length(*test)
        eq_(list(actual), expect)



from nose.tools import eq_
from ruamel.yaml import YAML

from moban.core.data_loader import merge


def test_simple_union():
    user = {"hi": "world"}
    default = {"world": "hi"}
    merged = merge(user, default)
    assert merged == {"hi": "world", "world": "hi"}


def test_simple_overlapping():
    user = {"hi": "world", "world": "hei"}
    default = {"world": "hi"}
    merged = merge(user, default)
    assert merged == {"hi": "world", "world": "hei"}


def test_two_level_merge():
    user = {"L1": {"L2": "World"}}
    default = {"L1": {"L2.1": "Hi"}}
    merged = merge(user, default)
    assert merged == {"L1": {"L2": "World", "L2.1": "Hi"}}


def test_two_level_conflict():
    user = {"L1": {"L2": "World"}}
    default = {"L1": {"L2": "Hi"}}
    merged = merge(user, default)
    assert merged == {"L1": {"L2": "World"}}


def test_three_level_conflict():
    user = {"L1": {"L2": {"L3": "World"}}}
    default = {"L1": {"L2": {"L3": "Hi"}}}
    merged = merge(user, default)
    assert merged == {"L1": {"L2": {"L3": "World"}}}


def test_merge_value_as_list():
    user = {"L1": ["a", "b"]}
    default = {"L1": ["c", "d"]}
    merged = merge(user, default)
    assert merged == {"L1": ["a", "b", "c", "d"]}


def test_merge_value_as_list_in_yaml():
    yaml = YAML(typ="rt")
    user = yaml.load(
        """
L1:
  - a
  - b
"""
    )
    default = yaml.load(
        """
L1:
  - c
  - d
"""
    )
    merged = merge(user, default)
    eq_(merged, {"L1": ["a", "b", "c", "d"]})



import json
import hashlib

from moban import constants, exceptions
from moban.externals import file_system


class HashStore:
    IGNORE_CACHE_FILE = False

    def __init__(self):
        self.cache_file = constants.DEFAULT_MOBAN_CACHE_FILE
        if (
            file_system.exists(self.cache_file)
            and self.IGNORE_CACHE_FILE is False
        ):
            with file_system.open_file(self.cache_file) as f:
                self.hashes = json.load(f)
        else:
            self.hashes = {}

    def is_file_changed(self, file_name, file_content, source_template):
        changed, with_permission = self._is_source_updated(
            file_name, file_content, source_template
        )

        if changed is False:
            target_hash = get_file_hash(
                file_name, with_permission=with_permission
            )
            if target_hash != self.hashes[file_name]:
                changed = True
        return changed

    def _is_source_updated(self, file_name, file_content, source_template):
        changed = True
        content = file_content
        with_permission = True
        try:
            content = _mix(
                file_content,
                oct(file_system.file_permissions(source_template)),
            )
        except exceptions.NoPermissionsNeeded:
            # HttpFs does not have getsyspath
            # zip, tar have no permission
            # win32 does not work
            with_permission = False
            pass
        content_hash = get_hash(content)
        if file_system.exists(file_name):
            if file_name in self.hashes:
                if content_hash == self.hashes[file_name]:
                    changed = False
        # else the dest file has not been created yet
        # so no need to get content hash at all
        if changed:
            self.hashes[file_name] = content_hash

        return changed, with_permission

    def save_hashes(self):
        with open(self.cache_file, "w") as f:
            json.dump(self.hashes, f)


HASH_STORE = HashStore()


def get_file_hash(afile, with_permission=True):
    content = file_system.read_bytes(afile)
    try:
        if with_permission:
            content = _mix(content, oct(file_system.file_permissions(afile)))
    except exceptions.NoPermissionsNeeded:
        # HttpFs does not have getsyspath
        # zip, tar have no permission
        # win32 does not work
        pass
    return get_hash(content)


def get_hash(content):
    md5 = hashlib.md5()
    md5.update(content)
    return md5.digest().decode("latin1")


def _mix(content, file_permissions_copy):
    file_permissions_copy = file_permissions_copy.encode("utf-8")
    return content + file_permissions_copy



from lml.plugin import PluginInfo
from ruamel.yaml import YAML

from moban import constants
from moban.externals.file_system import open_file


@PluginInfo(constants.DATA_LOADER_EXTENSION, tags=["yaml", "yml"])
def open_yaml(file_name):
    with open_file(file_name) as data_yaml:
        yaml = YAML(typ="rt")
        data = yaml.load(data_yaml)
        return data



import os
import sys
from shutil import copyfile

from mock import MagicMock, patch
from nose.tools import eq_, raises, assert_raises

import moban.exceptions as exceptions


class TestException:
    def setUp(self):
        self.moban_file = ".moban.yml"
        self.data_file = "data.yml"

    def tearDown(self):
        if os.path.exists(self.moban_file):
            os.unlink(self.moban_file)
        if os.path.exists(self.data_file):
            os.unlink(self.data_file)

    @raises(exceptions.MobanfileGrammarException)
    def test_handle_moban_file(self):
        copyfile(
            os.path.join("tests", "fixtures", ".moban-version-1234.yml"),
            self.moban_file,
        )
        import moban.main as main

        main.handle_moban_file(self.moban_file, {})

    def test_check_none(self):
        import moban.main as main
        from ruamel.yaml import YAML

        yaml = YAML()

        invalid_data = [
            """targets:
              - output:""",
            """configuration:
              template_dir:
                - cool-templates
                -""",
        ]

        valid_data = [
            """targets:
              - output: template""",
            """configuration:
              template_dir:
                - cool-templates
                - custom-templates""",
        ]

        for data in (yaml.load(d) for d in invalid_data):
            assert_raises(
                exceptions.MobanfileGrammarException,
                main.check_none,
                data,
                ".moban.yaml",
            )

        for data in (yaml.load(d) for d in valid_data):
            main.check_none(data, ".moban.yaml")

    @raises(exceptions.MobanfileGrammarException)
    def test_version_1_is_recognized(self):
        copyfile(
            os.path.join("tests", "fixtures", ".moban-version-1.0.yml"),
            self.moban_file,
        )
        copyfile(
            os.path.join("tests", "fixtures", ".moban-version-1.0.yml"),
            self.data_file,
        )
        import moban.main as main

        main.handle_moban_file(self.moban_file, {})

    @raises(SystemExit)
    @patch("os.path.exists")
    @patch("moban.main.handle_moban_file")
    @patch("moban.externals.reporter.report_error_message")
    def test_directory_not_found(
        self, fake_reporter, fake_moban_file, fake_file
    ):
        fake_file.return_value = True
        fake_moban_file.side_effect = exceptions.DirectoryNotFound
        fake_stdin = MagicMock(isatty=MagicMock(return_value=True))
        with patch.object(sys, "stdin", fake_stdin):
            with patch.object(sys, "argv", ["moban"]):
                from moban.main import main

                main()

    @raises(SystemExit)
    @patch("os.path.exists")
    @patch("moban.main.handle_moban_file")
    @patch("moban.externals.reporter.report_error_message")
    def test_unknown_protocol(self, fake_reporter, fake_moban_file, fake_file):
        fake_file.return_value = True
        fake_moban_file.side_effect = exceptions.UnsupportedPyFS2Protocol
        fake_stdin = MagicMock(isatty=MagicMock(return_value=True))
        with patch.object(sys, "stdin", fake_stdin):
            from moban.main import main

            with patch.object(sys, "argv", ["moban"]):
                main()

    @raises(SystemExit)
    @patch("os.path.exists")
    @patch("moban.main.handle_command_line")
    @patch("moban.externals.reporter.report_error_message")
    def test_unknown_protocol_at_command_line(
        self, fake_reporter, fake_moban_file, fake_file
    ):
        fake_file.return_value = False
        fake_moban_file.side_effect = exceptions.UnsupportedPyFS2Protocol
        fake_stdin = MagicMock(isatty=MagicMock(return_value=True))
        with patch.object(sys, "stdin", fake_stdin):
            from moban.main import main

            with patch.object(sys, "argv", ["moban"]):
                main()

    @raises(SystemExit)
    @patch("os.path.exists")
    @patch("moban.main.handle_moban_file")
    @patch("moban.externals.reporter.report_error_message")
    def test_no_third_party_engine(
        self, fake_reporter, fake_moban_file, fake_file
    ):
        fake_file.return_value = True
        fake_moban_file.side_effect = exceptions.NoThirdPartyEngine
        fake_stdin = MagicMock(isatty=MagicMock(return_value=True))
        with patch.object(sys, "stdin", fake_stdin):
            from moban.main import main

            with patch.object(sys, "argv", ["moban"]):
                main()

    @raises(SystemExit)
    @patch("os.path.exists")
    @patch("moban.main.handle_moban_file")
    @patch("moban.externals.reporter.report_error_message")
    def test_double_underscore_main(
        self, fake_reporter, fake_moban_file, fake_file
    ):
        fake_file.return_value = True
        fake_moban_file.side_effect = exceptions.DirectoryNotFound
        fake_stdin = MagicMock(isatty=MagicMock(return_value=True))
        with patch.object(sys, "stdin", fake_stdin):
            from moban.__main__ import main

            with patch.object(sys, "argv", ["moban"]):
                main()


class TestExitCodes:
    @raises(SystemExit)
    @patch("moban.main.handle_moban_file")
    @patch("moban.main.find_default_moban_file")
    def test_has_many_files_with_exit_code(
        self, fake_find_file, fake_moban_file
    ):
        fake_find_file.return_value = "abc"
        fake_moban_file.return_value = 1
        from moban.main import main

        with patch.object(sys, "argv", ["moban", "--exit-code"]):
            main()

    @raises(SystemExit)
    @patch("moban.main.handle_command_line")
    @patch("moban.main.find_default_moban_file")
    def test_handle_single_change_with_exit_code(
        self, fake_find_file, fake_command_line
    ):
        fake_find_file.return_value = None
        fake_command_line.return_value = 1
        from moban.main import main

        with patch.object(sys, "argv", ["moban", "--exit-code"]):
            main()

    @patch("moban.main.handle_moban_file")
    @patch("moban.main.find_default_moban_file")
    def test_has_many_files(self, fake_find_file, fake_moban_file):
        fake_find_file.return_value = "abc"
        fake_moban_file.return_value = 1
        from moban.main import main

        with patch.object(sys, "argv", ["moban"]):
            main()

    @patch("moban.main.handle_command_line")
    @patch("moban.main.find_default_moban_file")
    def test_handle_single_change(self, fake_find_file, fake_command_line):
        fake_find_file.return_value = None
        fake_command_line.return_value = 1
        from moban.main import main

        with patch.object(sys, "argv", ["moban"]):
            main()


class TestFinder:
    def setUp(self):
        self.patcher = patch("moban.externals.file_system.exists")
        self.fake_file_existence = self.patcher.start()
        self.fake_file_existence.__name__ = "fake"
        self.fake_file_existence.__module__ = "fake"

    def tearDown(self):
        self.patcher.stop()

    def test_moban_yml(self):
        self.fake_file_existence.return_value = True
        from moban.main import find_default_moban_file

        actual = find_default_moban_file()
        eq_(".moban.yml", actual)

    def test_moban_yaml(self):
        self.fake_file_existence.side_effect = [False, True]
        from moban.main import find_default_moban_file

        actual = find_default_moban_file()
        eq_(".moban.yaml", actual)

    def test_no_moban_file(self):
        self.fake_file_existence.side_effect = [False, False]
        from moban.main import find_default_moban_file

        actual = find_default_moban_file()
        assert actual is None



import os
import sys
from unittest.mock import patch

from nose import SkipTest

from moban.externals import file_system
from moban.exceptions import NoPermissionsNeeded
from moban.core.hashstore import HashStore, get_file_hash


class TestHashStore:
    def setUp(self):
        self.source_template = file_system.path_join(
            "tests", "fixtures", "a.jj2"
        )
        self.fixture = (
            "test.out",
            "test content".encode("utf-8"),
            self.source_template,
        )
        self.file_hash = get_file_hash(self.source_template)

    def tearDown(self):
        if os.path.exists(".moban.hashes"):
            os.unlink(".moban.hashes")

    def test_simple_use_case(self):
        hs = HashStore()
        flag = hs.is_file_changed(*self.fixture)

        hs.save_hashes()
        assert flag is True

    @patch("moban.core.hashstore.file_system.file_permissions")
    def test_permission_check_failed(self, fake):
        """
        when system permission fails, both source hash and
        target hash shall not use permission.
        """
        fake.side_effect = [NoPermissionsNeeded()]
        hs = HashStore()
        flag = hs.is_file_changed(*self.fixture)

        assert hs.hashes["test.out"] != self.file_hash
        hs.save_hashes()
        assert flag is True

    def test_dest_file_does_not_exist(self):
        hs = HashStore()
        flag = hs.is_file_changed(*self.fixture)
        hs.save_hashes()
        hs2 = HashStore()
        flag = hs2.is_file_changed(*self.fixture)
        assert flag is True

    def test_dest_file_exist(self):
        hs = HashStore()
        flag = hs.is_file_changed(*self.fixture)
        if flag:
            with open(self.fixture[0], "wb") as f:
                f.write(self.fixture[1])
        hs.save_hashes()
        hs2 = HashStore()
        flag = hs2.is_file_changed(*self.fixture)
        assert flag is False
        hs2.save_hashes()
        os.unlink(self.fixture[0])

    def test_dest_file_changed(self):
        """
        The situation is:

        moban once
        then update the generated file
        moban again, and the generated file should be detected
        and get templated
        """
        hs = HashStore()
        flag = hs.is_file_changed(*self.fixture)
        if flag:
            with open(self.fixture[0], "wb") as f:
                f.write(self.fixture[1])
        hs.save_hashes()
        # no change
        hs2 = HashStore()
        flag = hs2.is_file_changed(*self.fixture)
        assert flag is False
        hs2.save_hashes()
        # now let update the generated file
        hs3 = HashStore()
        with open(self.fixture[0], "w") as f:
            f.write("hey changed")
        flag = hs3.is_file_changed(*self.fixture)
        assert flag is True
        hs3.save_hashes()
        os.unlink(self.fixture[0])

    def test_dest_file_file_permision_changed(self):
        """
        Save as above, but this time,
        the generated file had file permision change
        """
        if sys.platform == "win32":
            raise SkipTest("No file permission check on windows")
        hs = HashStore()
        flag = hs.is_file_changed(*self.fixture)
        if flag:
            with open(self.fixture[0], "wb") as f:
                f.write(self.fixture[1])
        hs.save_hashes()
        # no change
        hs2 = HashStore()
        flag = hs2.is_file_changed(*self.fixture)
        assert flag is False
        hs2.save_hashes()
        # now let change file permision of generated file
        hs3 = HashStore()
        os.chmod(self.fixture[0], 0o766)
        flag = hs3.is_file_changed(*self.fixture)
        assert flag is True
        hs3.save_hashes()
        os.unlink(self.fixture[0])



import logging

from moban import constants
from moban.externals import reporter, file_system
from .store import STORE

LOG = logging.getLogger(__name__)


def handle_template(template_file, output, template_dirs):
    LOG.info(f"handling {template_file}")

    template_file = template_file
    multi_fs = file_system.get_multi_fs(template_dirs)
    if template_file.endswith("**"):
        source_dir = template_file[:-3]
        _, fs = multi_fs.which(source_dir)
        if fs:
            yield from _listing_directory_files_recusively(
                fs, source_dir, output
            )
        else:
            reporter.report_error_message(f"{template_file} cannot be found")
    else:
        if STORE.look_up_by_output.get(template_file) is None:
            _, fs = multi_fs.which(template_file)
            if fs is None:
                reporter.report_error_message(
                    f"{template_file} cannot be found"
                )
            elif fs.isdir(template_file):
                yield from _list_dir_files(fs, template_file, output)
            else:
                yield _create_a_single_target(template_file, output)
        else:
            # when template_file is not found, it means
            it_is_generated_by_moban = template_file
            STORE.intermediate_targets.append(it_is_generated_by_moban)
            yield _create_a_single_target(template_file, output)


def _list_dir_files(fs, source, dest):
    for file_name in fs.listdir(source):
        # please note jinja2 does NOT like windows path
        # hence the following statement looks like cross platform
        #  src_file_under_dir = os.path.join(source, file_name)
        # but actually it breaks windows instead.
        src_file_under_dir = f"{source}/{file_name}"
        if fs.isfile(src_file_under_dir):
            dest_file_under_dir = f"{dest}/{file_name}"
            template_type = _get_template_type(src_file_under_dir)
            yield (src_file_under_dir, dest_file_under_dir, template_type)


def _listing_directory_files_recusively(fs, source, dest):
    for file_name in fs.listdir(source):
        src_file_under_dir = f"{source}/{file_name}"
        dest_file_under_dir = f"{dest}/{file_name}"
        if fs.isfile(src_file_under_dir):
            template_type = _get_template_type(src_file_under_dir)
            yield (src_file_under_dir, dest_file_under_dir, template_type)
        elif fs.isdir(src_file_under_dir):
            yield from _listing_directory_files_recusively(
                fs, src_file_under_dir, dest_file_under_dir
            )


def _create_a_single_target(template_file, output):
    if output == constants.TEMPLATE_DELETE + "!":
        template_type = constants.TEMPLATE_DELETE
    else:
        template_type = _get_template_type(template_file)
    # output.jj2: source.jj2 means 'copy'
    if template_type and output.endswith("." + template_type):
        LOG.info(
            f"template type switched to from {template_type} to "
            + constants.TEMPLATE_COPY
        )
        template_type = constants.TEMPLATE_COPY
    return (template_file, output, template_type)


def _get_template_type(template_file):
    _, extension = file_system.path_splitext(template_file)
    if extension:
        template_type = extension[1:]
    else:
        template_type = None
    return template_type



import os
import copy
import logging

from moban import constants, exceptions
from moban.core import utils
from moban.program_options import OPTIONS
from moban.core.data_loader import merge, load_data

LOG = logging.getLogger(__name__)
MESSAGE_USING_ENV_VARS = "Attempting to use environment vars as data..."


class Context(object):
    def __init__(self, context_dirs):
        utils.verify_the_existence_of_directories(context_dirs)
        self.context_dirs = context_dirs
        self.__cached_environ_variables = dict(
            (key, os.environ[key]) for key in os.environ
        )

    def get_data(self, file_name):
        custom_data = copy.deepcopy(OPTIONS[constants.CLI_DICT])
        try:
            data = load_data(self.context_dirs, file_name)
            merge(custom_data, data)
            merge(custom_data, self.__cached_environ_variables)
            return custom_data
        except (IOError, exceptions.IncorrectDataInput) as exception:
            # If data file doesn't exist:
            # 1. Alert the user of their (potential) mistake
            # 2. Attempt to use environment vars as data
            LOG.warn(str(exception))
            LOG.info(MESSAGE_USING_ENV_VARS)
            merge(custom_data, self.__cached_environ_variables)
            return custom_data



import fs.path
from mock import patch
from nose.tools import eq_, raises

from moban.deprecated import GitRequire
from moban.exceptions import NoGitCommand
from moban.deprecated.repo import (
    git_clone,
    get_repo_name,
    get_moban_home,
    make_sure_git_is_available,
)


@patch("appdirs.user_cache_dir", return_value="root")
@patch("moban.utils.mkdir_p")
@patch("moban.file_system.exists")
@patch("git.Repo", autospec=True)
class TestGitFunctions:
    def setUp(self):
        self.repo_name = "repoA"
        self.repo = "https://github.com/my/" + self.repo_name
        self.require = GitRequire(git_url=self.repo)
        self.require_with_submodule = GitRequire(
            git_url=self.repo, submodule=True
        )
        self.require_with_branch = GitRequire(
            git_url=self.repo, branch="ghpages"
        )
        self.require_with_reference = GitRequire(
            git_url=self.repo, reference="a-commit-reference"
        )
        self.expected_local_repo_path = fs.path.join(
            "root", "repos", self.repo_name
        )

    def test_checkout_new(self, fake_repo, local_folder_exists, *_):
        local_folder_exists.return_value = False
        git_clone([self.require])
        fake_repo.clone_from.assert_called_with(
            self.repo,
            self.expected_local_repo_path,
            single_branch=True,
            depth=2,
        )
        repo = fake_repo.return_value
        eq_(repo.git.submodule.called, False)

    def test_checkout_new_with_submodules(
        self, fake_repo, local_folder_exists, *_
    ):
        local_folder_exists.return_value = False
        git_clone([self.require_with_submodule])
        fake_repo.clone_from.assert_called_with(
            self.repo,
            self.expected_local_repo_path,
            single_branch=True,
            depth=2,
        )
        repo = fake_repo.clone_from.return_value
        repo.git.submodule.assert_called_with("update", "--init")

    def test_git_update(self, fake_repo, local_folder_exists, *_):
        local_folder_exists.return_value = True
        git_clone([self.require])
        fake_repo.assert_called_with(self.expected_local_repo_path)
        repo = fake_repo.return_value
        repo.git.pull.assert_called()

    def test_git_update_with_submodules(
        self, fake_repo, local_folder_exists, *_
    ):
        local_folder_exists.return_value = True
        git_clone([self.require_with_submodule])
        fake_repo.assert_called_with(self.expected_local_repo_path)
        repo = fake_repo.return_value
        repo.git.submodule.assert_called_with("update")

    def test_checkout_new_with_branch(
        self, fake_repo, local_folder_exists, *_
    ):
        local_folder_exists.return_value = False
        git_clone([self.require_with_branch])
        fake_repo.clone_from.assert_called_with(
            self.repo,
            self.expected_local_repo_path,
            branch="ghpages",
            single_branch=True,
            depth=2,
        )
        repo = fake_repo.return_value
        eq_(repo.git.submodule.called, False)

    def test_update_existing_with_branch_parameter(
        self, fake_repo, local_folder_exists, *_
    ):
        local_folder_exists.return_value = True
        git_clone([self.require_with_branch])
        repo = fake_repo.return_value
        repo.git.checkout.assert_called_with("ghpages")

    def test_checkout_new_with_reference(
        self, fake_repo, local_folder_exists, *_
    ):
        local_folder_exists.return_value = False
        git_clone([self.require_with_reference])
        fake_repo.clone_from.assert_called_with(
            self.repo,
            self.expected_local_repo_path,
            reference="a-commit-reference",
            single_branch=True,
            depth=2,
        )
        repo = fake_repo.return_value
        eq_(repo.git.submodule.called, False)

    def test_update_existing_with_reference_parameter(
        self, fake_repo, local_folder_exists, *_
    ):
        local_folder_exists.return_value = True
        git_clone([self.require_with_reference])
        repo = fake_repo.return_value
        repo.git.checkout.assert_called_with("a-commit-reference")


def test_get_repo_name():
    repos = [
        "https://github.com/repo-abc-def/repo",
        "https://github.com/abc/repo",
        "https://github.com/abc/repo.git",
        "https://github.com/abc/repo/",
        "git@github.com:abc/repo.git",
        "git@bitbucket.org:abc/repo.git",
        "git://github.com/abc/repo.git",
    ]
    actual = [get_repo_name(repo) for repo in repos]
    expected = ["repo"] * len(repos)
    eq_(expected, actual)


@patch("moban.reporter.report_error_message")
def test_get_repo_name_can_handle_invalid_url(fake_reporter):
    invalid_repo = "invalid"
    try:
        get_repo_name(invalid_repo)
    except Exception:
        fake_reporter.assert_called_with(
            'An invalid git url: "invalid" in mobanfile'
        )


@patch("appdirs.user_cache_dir", return_value="root")
def test_get_moban_home(_):
    actual = get_moban_home()
    eq_(fs.path.join("root", "repos"), actual)


@raises(NoGitCommand)
@patch("subprocess.check_output", side_effect=Exception)
def test_make_git_is_available(_):
    make_sure_git_is_available()



from lml.plugin import PluginInfo

from moban import constants


class JinjaFilter(PluginInfo):
    def __init__(self):
        super(JinjaFilter, self).__init__(constants.JINJA_FILTER_EXTENSION)

    def tags(self):
        yield self.cls.__name__


class JinjaTest(PluginInfo):
    def __init__(self, test_name=None):
        super(JinjaTest, self).__init__(constants.JINJA_TEST_EXTENSION)
        self.test_name = test_name

    def tags(self):
        if self.test_name:
            yield self.test_name
        else:
            yield self.cls.__name__


def jinja_tests(**keywords):
    for key, value in keywords.items():
        JinjaTest(key)(value)


def jinja_global(identifier, dict_obj):
    plugin = PluginInfo(constants.JINJA_GLOBALS_EXTENSION, tags=[identifier])
    plugin(dict_obj)



import os
import sys
import stat
import errno
import logging
from contextlib import contextmanager

import fs
import fs.path
from fs.multifs import MultiFS

from moban import exceptions

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse


LOG = logging.getLogger(__name__)

path_join = fs.path.join
path_splitext = fs.path.splitext


def url_join(path, path2):
    result = urlparse(path)
    if result.scheme and path.endswith(result.scheme):
        return f"{path}!/{path2}"
    else:
        return f"{path}/{path2}"


def log_fs_failure(function_in_this_module):
    def wrapper(*args, **kwds):
        try:
            return function_in_this_module(*args, **kwds)
        except fs.errors.CreateFailed:
            message = "Failed to open %s" % args[0]
            LOG.debug(message)
            raise exceptions.FileNotFound(args[0])
        except fs.opener.errors.UnsupportedProtocol as e:
            LOG.exception(e)
            raise exceptions.UnsupportedPyFS2Protocol(e)

    return wrapper


@log_fs_failure
@contextmanager
def open_fs(path):
    if is_zip_alike_url(path):
        zip_file, folder = url_split(path)
        the_fs = fs.open_fs(zip_file)
    else:
        the_fs = fs.open_fs(path)
    try:
        yield the_fs
    finally:
        the_fs.close()


@log_fs_failure
@contextmanager
def open_file(path):
    if is_zip_alike_url(path):
        zip_file, folder = url_split(path)
        the_fs = fs.open_fs(zip_file)
        f = the_fs.open(folder)
    else:
        dir_name = fs.path.dirname(path)
        the_file_name = fs.path.basename(path)
        the_fs = fs.open_fs(dir_name)
        f = the_fs.open(the_file_name)
    try:
        yield f
    finally:
        f.close()
        the_fs.close()


@log_fs_failure
@contextmanager
def open_binary_file(path):
    if is_zip_alike_url(path):
        zip_file, folder = url_split(path)
        the_fs = fs.open_fs(zip_file)
        f = the_fs.openbin(folder)
    else:
        dir_name = fs.path.dirname(path)
        the_file_name = fs.path.basename(path)
        the_fs = fs.open_fs(dir_name)
        f = the_fs.openbin(the_file_name)
    try:
        yield f
    finally:
        f.close()
        the_fs.close()


@log_fs_failure
def read_unicode(path):
    with open_file(path) as file_handle:
        return file_handle.read()


@log_fs_failure
def read_bytes(path):
    with open_binary_file(path) as file_handle:
        return file_handle.read()


read_binary = read_bytes
read_text = read_unicode


@log_fs_failure
def write_bytes(filename, bytes_content):
    if "://" in filename:
        zip_file, folder = url_split(filename)
        with fs.open_fs(zip_file, create=True) as the_fs:
            the_fs.writebytes(folder, bytes_content)
    else:
        dir_name = fs.path.dirname(filename)
        the_file_name = fs.path.basename(filename)
        with fs.open_fs(dir_name) as the_fs:
            the_fs.writebytes(the_file_name, bytes_content)


@log_fs_failure
def is_dir(path):
    folder_or_file, path = _path_split(path)
    with fs.open_fs(folder_or_file) as the_fs:
        return the_fs.isdir(path)


@log_fs_failure
def is_file(path):
    folder_or_file, path = _path_split(path)
    with fs.open_fs(folder_or_file) as the_fs:
        return the_fs.isfile(path)


def exists(path):
    if is_zip_alike_url(path):
        zip_file, folder = url_split(path)
        try:
            with fs.open_fs(zip_file) as the_fs:
                if folder:
                    return the_fs.exists(folder)
                return True
        except fs.errors.CreateFailed:
            return False
        except fs.opener.errors.UnsupportedProtocol as e:
            LOG.exception(e)
            raise exceptions.UnsupportedPyFS2Protocol(e)
    dir_name = fs.path.dirname(path)
    the_file_name = fs.path.basename(path)

    try:
        with fs.open_fs(dir_name) as the_fs:
            return the_fs.exists(the_file_name)
    except fs.errors.CreateFailed:
        return False
    except fs.opener.errors.UnsupportedProtocol as e:
        LOG.exception(e)
        raise exceptions.UnsupportedPyFS2Protocol(e)


@log_fs_failure
def list_dir(path):
    folder_or_file, path = _path_split(path)
    with fs.open_fs(folder_or_file) as the_fs:
        for file_name in the_fs.listdir(path):
            yield file_name


@log_fs_failure
def abspath(path):
    folder_or_file, path = _path_split(path)
    with fs.open_fs(folder_or_file) as the_fs:
        return the_fs.getsyspath(path)


@log_fs_failure
def fs_url(path):
    folder_or_file, path = _path_split(path)
    with fs.open_fs(folder_or_file) as the_fs:
        return the_fs.geturl(path, purpose="fs")


@log_fs_failure
def system_path(path):
    folder_or_file, path = _path_split(path)
    with fs.open_fs(folder_or_file) as the_fs:
        return the_fs.getsyspath(path)


def to_unicode(path):
    return path


def is_zip_alike_url(url):
    specs = ["zip://", "tar://"]
    for prefix in specs:
        if url.startswith(prefix):
            return True
    else:
        return False


def file_permissions_copy(source, dest):
    source_permissions = file_permissions(source)
    dest_permissions = file_permissions(dest)

    if source_permissions != dest_permissions:
        os.chmod(dest, source_permissions)


def file_permissions(url):
    if not exists(url):
        raise exceptions.FileNotFound(url)
    if sys.platform == "win32":
        raise exceptions.NoPermissionsNeeded()
    elif is_zip_alike_url(url):
        raise exceptions.NoPermissionsNeeded()
    else:
        try:
            unix_path = system_path(url)
            return stat.S_IMODE(os.stat(unix_path).st_mode)
        except fs.errors.NoSysPath:
            raise exceptions.NoPermissionsNeeded()


def url_split(url):
    result = urlparse(url)

    if url.endswith(result.scheme):
        url_to_file = url
        path = None
    else:
        url_to_file, path = url.split("!/")

    return url_to_file, path


def _path_split(url_or_path):
    if is_zip_alike_url(url_or_path):
        return url_split(url_or_path)
    else:
        return fs.path.dirname(url_or_path), fs.path.basename(url_or_path)


def get_multi_fs(directories):
    filesystem = MultiFS()
    for directory in directories:
        filesystem.add_fs(directory, fs.open_fs(directory))
    return filesystem


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise



import sys
from functools import wraps

from moban import constants
from moban.core import plugins
from moban.externals import reporter, file_system
from moban.deprecated import repo
from moban.deprecated.repo import git_clone
from moban.deprecated.library import LIBRARIES

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse


KNOWN_DOMAIN_FOR_GIT = ["github.com", "gitlab.com", "bitbucket.com"]


def deprecated(message):
    def tags_decorator(func):
        @wraps(func)
        def func_wrapper(*args, **kwds):
            reporter.report_warning_message(message)
            return func(*args, **kwds)

        return func_wrapper

    return tags_decorator


class GitRequire(object):
    def __init__(
        self, git_url=None, branch=None, submodule=False, reference=None
    ):
        self.git_url = git_url
        self.submodule = submodule
        self.branch = branch
        self.reference = reference

    def clone_params(self):
        clone_params = {
            "single_branch": True,
            "depth": constants.DEFAULT_CLONE_DEPTH,
        }
        if self.branch is not None:
            clone_params["branch"] = self.branch
        elif self.reference is not None:
            clone_params["reference"] = self.reference
        return clone_params

    def __eq__(self, other):
        return (
            self.git_url == other.git_url
            and self.submodule == other.submodule
            and self.branch == other.branch
            and self.reference == other.reference
        )

    def __repr__(self):
        return "%s,%s,%s" % (self.git_url, self.branch, self.submodule)


def pip_install(packages):
    import subprocess

    subprocess.check_call(
        [sys.executable, "-m", "pip", "install", " ".join(packages)],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )


@deprecated(constants.MESSAGE_DEPRECATE_MOBAN_NOTATION_SINCE_0_6_0)
def deprecated_moban_path_notation(directory):
    translated_directory = None
    library_or_repo_name, relative_path = directory.split(":")
    potential_repo_path = file_system.path_join(
        repo.get_moban_home(), library_or_repo_name
    )
    if file_system.exists(potential_repo_path):
        # expand repo template path
        if relative_path:
            translated_directory = file_system.path_join(
                potential_repo_path, relative_path
            )
        else:
            translated_directory = potential_repo_path
    else:
        # expand pypi template path
        library_path = LIBRARIES.resource_path_of(library_or_repo_name)
        if relative_path:
            translated_directory = file_system.path_join(
                library_path, relative_path
            )
        else:
            translated_directory = library_path
    return translated_directory


@deprecated(constants.MESSAGE_DEPRECATE_REQUIRES_SINCE_0_6_0)
def handle_requires(requires):
    pypi_pkgs = []
    git_repos = []
    for require in requires:
        if isinstance(require, dict):
            require_type = require.get(constants.REQUIRE_TYPE, "")
            if require_type.upper() == constants.GIT_REQUIRE:
                git_require = GitRequire(
                    git_url=require.get(constants.GIT_URL),
                    branch=require.get(constants.GIT_BRANCH),
                    reference=require.get(constants.GIT_REFERENCE),
                    submodule=require.get(constants.GIT_HAS_SUBMODULE, False),
                )

                git_repos.append(git_require)
            elif require_type.upper() == constants.PYPI_REQUIRE:
                pypi_pkgs.append(require.get(constants.PYPI_PACKAGE_NAME))
        else:
            if is_repo(require):
                git_repos.append(GitRequire(require))
            else:
                pypi_pkgs.append(require)
    if pypi_pkgs:
        pip_install(pypi_pkgs)
        plugins.make_sure_all_pkg_are_loaded()
    if git_repos:
        git_clone(git_repos)


def is_repo(require):
    result = urlparse(require)
    return result.scheme != "" and result.netloc in KNOWN_DOMAIN_FOR_GIT


@deprecated(constants.MESSAGE_DEPRECATE_COPY_SINCE_0_4_0)
def handle_copy(merged_options, copy_config):
    copy_targets = []
    for (dest, src) in _iterate_list_of_dicts(copy_config):
        copy_targets.append(
            {
                constants.LABEL_TEMPLATE: src,
                constants.LABEL_OUTPUT: dest,
                constants.LABEL_TEMPLATE_TYPE: constants.TEMPLATE_COPY,
            }
        )
    return copy_targets


def _iterate_list_of_dicts(list_of_dict):
    for adict in list_of_dict:
        for key, value in adict.items():
            yield (key, value)



from lml.loader import scan_plugins_regex

from moban import constants

BUILTIN_EXENSIONS = [
    "moban.plugins.jinja2.engine",
    "moban.plugins.yaml_loader",
    "moban.plugins.json_loader",
    "moban.plugins.copy",
    "moban.plugins.delete",
]


def make_sure_all_pkg_are_loaded():
    scan_plugins_regex(constants.MOBAN_ALL, "moban", None, BUILTIN_EXENSIONS)



import sys
import subprocess

from moban import constants, exceptions
from moban.externals import reporter, file_system


def git_clone(requires):
    from git import Repo

    if sys.platform != "win32":
        # Unfortunately for windows user, the following function
        # needs shell=True, which expose security risk. I would
        # rather not to trade it with its marginal benefit
        make_sure_git_is_available()

    moban_home = get_moban_home()
    file_system.mkdir_p(moban_home)

    for require in requires:
        repo_name = get_repo_name(require.git_url)
        local_repo_folder = file_system.path_join(moban_home, repo_name)
        if file_system.exists(local_repo_folder):
            reporter.report_git_pull(repo_name)
            repo = Repo(local_repo_folder)
            repo.git.pull()
            if require.reference:
                repo.git.checkout(require.reference)
            elif require.branch:
                repo.git.checkout(require.branch)
            if require.submodule:
                reporter.report_info_message("updating submodule")
                repo.git.submodule("update")
        else:
            reporter.report_git_clone(require.git_url)
            repo = Repo.clone_from(
                require.git_url, local_repo_folder, **require.clone_params()
            )
            if require.submodule:
                reporter.report_info_message("checking out submodule")
                repo.git.submodule("update", "--init")


def get_repo_name(repo_url):
    import giturlparse
    from giturlparse.parser import ParserError

    try:
        repo = giturlparse.parse(repo_url.rstrip("/"))
        return repo.name
    except ParserError:
        reporter.report_error_message(
            constants.MESSAGE_INVALID_GIT_URL % repo_url
        )
        raise


def get_moban_home():
    from appdirs import user_cache_dir

    home_dir = user_cache_dir(appname=constants.PROGRAM_NAME)
    return file_system.path_join(home_dir, constants.MOBAN_REPOS_DIR_NAME)


def make_sure_git_is_available():
    try:
        subprocess.check_output(["git", "--help"])
    except Exception:
        raise exceptions.NoGitCommand("Please install git command")



import os

from nose.tools import eq_

from moban import file_system
from moban.jinja2.engine import Engine


def test_jinja2_template():
    path = os.path.join("tests", "fixtures", "jinja_tests")
    fs = file_system.get_multi_fs([path])
    engine = Engine(fs)
    template = engine.get_template("file_tests.template")
    data = dict(test="here")
    result = engine.apply_template(template, data, None)
    expected = "yes\nhere"
    eq_(expected, result)


def test_jinja2_template_string():
    path = os.path.join("tests", "fixtures", "jinja_tests")
    fs = file_system.get_multi_fs([path])
    engine = Engine(fs)
    template = engine.get_template_from_string("{{test}}")
    data = dict(test="here")
    result = engine.apply_template(template, data, None)
    expected = "here"
    eq_(expected, result)



import os

import fs
import fs.path

from moban.externals import file_system


class BufferedWriter(object):
    def __init__(self):
        self.fs_list = {}

    def write_file_out(self, filename, content):
        if file_system.is_zip_alike_url(filename):
            self.write_file_out_to_zip(filename, content)
        else:
            write_file_out(filename, content)

    def write_file_out_to_zip(self, filename, content):
        zip_file, file_name = file_system.url_split(filename)
        if zip_file not in self.fs_list:
            self.fs_list[zip_file] = fs.open_fs(zip_file, create=True)
        base_dirs = fs.path.dirname(file_name)
        if not self.fs_list[zip_file].exists(base_dirs):
            self.fs_list[zip_file].makedirs(base_dirs)
        self.fs_list[zip_file].writebytes(file_name, content)

    def close(self):
        for fsx in self.fs_list.values():
            fsx.close()


def write_file_out(filename, content):
    if not file_system.is_zip_alike_url(filename):
        dest_folder = os.path.dirname(filename)
        if dest_folder:
            file_system.mkdir_p(dest_folder)

    file_system.write_bytes(filename, content)



class DirectoryNotFound(Exception):
    pass


class FileNotFound(Exception):
    pass


class NoThirdPartyEngine(Exception):
    pass


class MobanfileGrammarException(Exception):
    pass


class NoTemplate(Exception):
    pass


class IncorrectDataInput(Exception):
    pass


class GroupTargetNotFound(Exception):
    pass


class NoGitCommand(Exception):
    pass


class UnsupportedPyFS2Protocol(Exception):
    pass


class NoPermissionsNeeded(Exception):
    pass


class SingleHTTPURLConstraint(Exception):
    pass
